/*
Copyright 2009 Antus ant@symons.net.au

This file is part of Antusprom.

Antusprom is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Antusprom is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Antusprom.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <windows.h>
#include "stdafx.h"

#define VMAJ 0
#define VMIN 7

HANDLE hSerialECU;
HANDLE hSerialTP;
HANDLE hSerialSP;
HANDLE ghMutex;
unsigned char emurom[0xFFFF]= {0};
int ecu=1; // is the ecu connected?
int	echo=1; // expect cable echo?

unsigned char delcochecksum(unsigned char *buffer, DWORD length)
{
	DWORD i;
	unsigned char cs=0;

	for (i=0; i<length; i++) cs += buffer[i];
	cs=0x100-cs;

	return cs;
}
unsigned char *readEcuMemory(HANDLE hSerial, DWORD length, int address)
{
	unsigned char *buffer;
	unsigned char checksum;
	unsigned char packetbuffer[0xFF];
	unsigned char request[6] = {0xF4, 0x58, 0x02, 0x00, 0x00, 0x00};
	DWORD dwBytesCount = 0;
	DWORD i;
	DWORD offset;

	buffer = (unsigned char*) malloc(length+1);

	if (buffer == NULL)	{
		return buffer;
	}

	for (offset=0; offset<length; offset+=64)	{
		request[3] = ((address+offset) & 0xFF00)>>8;
		request[4] = (address+offset) & 0xFF;
		request[5] = delcochecksum(request, 5);
		do	{
			printf ("Reading 0x%04X\n", address);
			printf ("Sending 0x%02X%02X%02X%02X%02X%02X\n", request[0], request[1], request[2], request[3], request[4], request[5]);
			
			FlushFileBuffers(hSerial);
			
			if (!WriteFile(hSerial, request, sizeof(request), &dwBytesCount, NULL))	{
				printf("error writing to ecu - retry\n");
				continue;
			}
	
			FlushFileBuffers(hSerial);
			printf ("sent %d bytes\n", dwBytesCount);
			
			// read back cable echo
			if (echo == 1)	{
				printf("reading echo\n");
				ReadFile(hSerial, packetbuffer, dwBytesCount, &dwBytesCount, NULL);
			}
	
			if (!ReadFile(hSerial, packetbuffer, 80, &dwBytesCount, NULL))	{
				printf("error reading from ecu - retry\n");
				continue;
			}
			printf("Read %d bytes\n",dwBytesCount);
			if (dwBytesCount != 68 )	{
				printf ("fail, retry\n");
				continue;
			}
			
			checksum=0;
			for (i=0; i<dwBytesCount-1; i++)	{
				checksum+=packetbuffer[i];
			}

			checksum=0x100-(checksum & 0xFF);
			printf("checksum=0x%X vs received 0x%X\n", checksum, packetbuffer[dwBytesCount-1]);
			
			if (checksum != packetbuffer[dwBytesCount-1])	{
				printf ("fail, retry");
				continue;
			}
		}	while (dwBytesCount != 68);
			
		for (i=offset; i<offset+64, i<length; i++)	{
			//printf("copying packet byte %d of 0x%02X to buffer position %d\n", 3+i-offset, packetbuffer[3+i-offset], i);
			buffer[i]=packetbuffer[3+i-offset];
		}
	}

	printf("complete, returning buffer:\n");
	for (i=0; i<length; i++)	{
		printf ("%02X", buffer[i]);
	}
	printf ("\ndone\n");
	return buffer;
}

DWORD writeEcuMemory(HANDLE hSerial, DWORD length, int address, unsigned char *data)
{
	unsigned char request[5+8+1]; // header + max length + checksum
	unsigned char readbuf[5];
	DWORD dwBytesCount = 0;
	DWORD i,good;
	DWORD offset;
	unsigned char thislength;

	printf("write ecu memory says: trying to write %02X%02X%02X%02X%02X%02X%02X%02X\n",data[0],data[1],data[2],data[3],data[4],data[5],data[6],data[7]);
					
	for (offset=0; offset<length; offset+=8)	{
		// length
		if (length >= offset+8)	{
			thislength=0x08;
		}
		else	{
			thislength=length-offset;
		}
		printf("this request is %d bytes\n", thislength);
		request[0] = 0xF4;
		request[1] = 0x55+3+thislength;
		request[2] = 0x0B; //Mode 0x0B write packet
		// memory location 
		request[3] = ((address+offset) & 0xFF00)>>8;
		request[4] = (address+offset) & 0xFF;

		good=0;
		while (good!=1)	{
			printf ("Writing %d bytes to 0x%04X ", thislength, address+offset);
			for (i=offset; i<thislength; i++)	{ // stop at the end, or the next 8 byte boundry
				//printf("copying 0x%02X to write buffer\n",data[i]);
				request[5+i] = data[i];
			}

			request[5+thislength]=delcochecksum(request, 5+thislength); // generate checksum
			printf ("with delco checksum 0x%02X\n", request[5+thislength]);

			printf ("Writing: ");
			for (i=0; i<5+thislength+1; i++)	{
				printf("%02X",request[i]);
			}
			printf("\n");
			FlushFileBuffers(hSerial);
			if (!WriteFile(hSerial, request, 5+thislength+1, &dwBytesCount, NULL))	{
				printf("error writing to ecu\n\n");
				continue;
			}
			
			// read back echo
			if (echo == 1)	{
				printf("reading echo\n");
				ReadFile(hSerial, readbuf, dwBytesCount, &dwBytesCount, NULL);
			}

			if (!ReadFile(hSerial, readbuf, sizeof(readbuf), &dwBytesCount, NULL) || dwBytesCount != 4)	{
				printf("error reading from ecu - retry\n");
				continue;
			}
			printf("read %02X%02X%02X%02X\n", readbuf[0], readbuf[1], readbuf[2], readbuf[3]);
			// check return code
			if (readbuf[0] != 0xF4 || readbuf[1] != 0x56 || readbuf[2] != 0x0B )	{
				printf("bad response from ecu - retry\n");
				continue;
			}
			
			printf ("Write OK\n");
			good=1;
		}
	}
	return 1;
}

int setup_other_comm(_TCHAR *port, HANDLE &h)
{
	DCB dcbSerialParams = {0};
	COMMTIMEOUTS timeouts={0};

	// Now open our half of the com0com link to tunerproRT
	h = CreateFile(port,
					GENERIC_READ | GENERIC_WRITE,
					0,
					0,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,
					0);
	if (h==INVALID_HANDLE_VALUE)	{
		if (GetLastError()==ERROR_FILE_NOT_FOUND)	{
			printf("%S not found\n\n", port);
			exit(2);
		}
		printf("%S unknown error %d\n\n", port, GetLastError());
		exit(2);
	}
	
	printf("Serial port %S opened\n", port);

	if (!GetCommState(h, &dcbSerialParams))	{
		printf("error getting com port state\n\n");
		exit(3);
	}

	dcbSerialParams.BaudRate=115200;
	dcbSerialParams.Parity=NOPARITY;
	dcbSerialParams.ByteSize=8;
	dcbSerialParams.StopBits=ONESTOPBIT;
	dcbSerialParams.fBinary=TRUE;
	dcbSerialParams.fOutxCtsFlow=FALSE;
	dcbSerialParams.fOutxDsrFlow=FALSE;
	dcbSerialParams.fDtrControl=DTR_CONTROL_DISABLE;
	dcbSerialParams.fRtsControl=RTS_CONTROL_DISABLE;

	if (!SetCommState(h, &dcbSerialParams))	{
		printf("error setting com port state\n\n");
		exit(4);	
	}
	timeouts.ReadIntervalTimeout=50;
	timeouts.ReadTotalTimeoutConstant=50;
	timeouts.ReadTotalTimeoutMultiplier=1;
	timeouts.WriteTotalTimeoutConstant=50;
	timeouts.WriteTotalTimeoutMultiplier=1;
	if (!SetCommTimeouts(h, &timeouts))	{
		printf("error setting com port timeouts\n\n");
		exit(5);	
	}
	return 1;
}

int setup_ecu_comm(_TCHAR *port)
{
	DCB dcbSerialParams = {0};
	COMMTIMEOUTS timeouts={0};
	int i;
	unsigned char *tempbuffer;
	
	hSerialECU = CreateFile(port,
							GENERIC_READ | GENERIC_WRITE,
							0,
							0,
							OPEN_EXISTING,
							FILE_ATTRIBUTE_NORMAL,
							0);
	if (hSerialECU==INVALID_HANDLE_VALUE)	{
		if (GetLastError()==ERROR_FILE_NOT_FOUND)	{
			printf("%S not found\n\n", port);
		} 
		else {
			printf("%S unknown error\n\n", port);
		}
		ecu=0;
	}
	
	printf("Serial port %S opened\n", port);

	if (!GetCommState(hSerialECU, &dcbSerialParams))	{
		printf("error getting com port state\n\n");
	}

	dcbSerialParams.BaudRate=8192;
	dcbSerialParams.ByteSize=8;
	dcbSerialParams.StopBits=ONESTOPBIT;
	dcbSerialParams.Parity=NOPARITY;
	dcbSerialParams.fBinary=TRUE;
	dcbSerialParams.fOutxCtsFlow=FALSE;
	dcbSerialParams.fOutxDsrFlow=FALSE;
	dcbSerialParams.fDtrControl=DTR_CONTROL_DISABLE;
	dcbSerialParams.fRtsControl=RTS_CONTROL_DISABLE;

	if (!SetCommState(hSerialECU, &dcbSerialParams))	{
		printf("error setting com port state\n\n");
	}

	printf ("ECU com port opened\n");
	
	timeouts.ReadIntervalTimeout=-1;
	timeouts.ReadTotalTimeoutConstant=70;
	timeouts.ReadTotalTimeoutMultiplier=1;
	timeouts.WriteTotalTimeoutConstant=50;
	timeouts.WriteTotalTimeoutMultiplier=1;
	if (!SetCommTimeouts(hSerialECU, &timeouts))	{
		printf("error setting com port timeouts\n\n");
	}
	
	//ecu=0;
	if (ecu)	{
		printf ("ECU com port init OK\n\n");
		printf ("Copying current ROM to RAM\n\n");
		// now grab the rom
		for (i=0x8000; i<0xFFFF; i+=64)	{
			tempbuffer=readEcuMemory(hSerialECU, 64, i);
			if ( tempbuffer != NULL)	{
				memcpy(&emurom[i], tempbuffer, 64);
				free(tempbuffer);
			}
		}
		// now flag all bytes as 'clean'
		memset(emurom, 0, 0x8000);
	}
	else	{
		printf ("running in test mode (no ecu comms)\n");
		printf ("not copying ecu rom to pc ram\n");
	}
	return 1;
}

DWORD WINAPI WriteThread(LPVOID emuromp)
{
	int i, j, k, d=-1;
	unsigned char *emurom=(unsigned char *)emuromp;

	for(;;)	{
		do	{
			// always upload 8 bytes, and stop 8 bytes from the end. 
			// will need to change if flashing new reset vector etc
			for (i=0, d=0; i<0x8000-8; i++)	{ 
				k=i+0x8000;
				if (emurom[i]==1)	{
					d++;
					for (j=0; j<8; j++)	{
						emurom[i+j]=0;
					}
					WaitForSingleObject( ghMutex, INFINITE);
					printf("write thread says: trying to write %02X%02X%02X%02X%02X%02X%02X%02X\n",emurom[k],emurom[k+1],emurom[k+2],emurom[k+3],emurom[k+4],emurom[k+5],emurom[k+6],emurom[k+7]);
					if (!writeEcuMemory(hSerialECU, 8, k, &emurom[k])){
						printf("Failed write to 0x%4X\n", k);
						for (j=0; j<8; j++)	{
							emurom[i+j]=1;
						}
					}
					ReleaseMutex(ghMutex);
				}
			} 
			if (d) printf("%d blocks of 8 uploaded to ECU\n", d);
		} while (d != 0);
	Sleep(1000);
	}
	return 0;
}

DWORD WINAPI input_handler_thread(LPVOID which)
{
	HANDLE *h= (HANDLE *)which;

	// comms stuff
	unsigned char packetbuffer[0x280];
	unsigned char outbuffer[0x280];
	DWORD dwBytesCount;
	unsigned char autopromid[]={0x02, 0x05, 'A'};
	unsigned char okay[]={'O'};

	// emu stuff
	DWORD i, len, address;
	unsigned char cs;

	while (1)	{
		if (ReadFile(h, packetbuffer, sizeof(packetbuffer), &dwBytesCount, NULL)>0 && dwBytesCount !=0 )	{
			len=dwBytesCount;
			if (packetbuffer[0] >=  33 && packetbuffer[0] <= 126 )	{
				printf("Read command %c, size %d\n", packetbuffer[0], dwBytesCount);
			}
			else	{
				printf("Read command 0x%02X, size %d\n", packetbuffer[0], dwBytesCount);
			}
			switch (packetbuffer[0])	{
				case 'V':	WaitForSingleObject( ghMutex, INFINITE);
							printf("Version requested, sending V2.5 AutoProm when locked\n");
							if (WriteFile(h, autopromid, sizeof(autopromid), &dwBytesCount, NULL))	{
								printf ("tunerproRT init OK\n");	
							}
							else{
								printf ("tunerproRT init Failed\n");
							}
							ReleaseMutex(ghMutex);
							break;
				case 'r':	// backwards compatible r, used by older tunerpro, but same as R
				case 'R':	len=packetbuffer[1]==0?256:packetbuffer[1];
							address=(packetbuffer[2]<<8)+packetbuffer[3];
							printf("Read request for %d bytes from location 0x%04X\n", len, address);
							//tempbuffer=readEcuMemory(hSerialECU, len, address);
							for (cs=0, i=0; i<len; i++)	{
								//outbuffer[i]=tempbuffer[i];
								outbuffer[i]=emurom[address+i];
								cs+=outbuffer[i];
							}
							printf ("checksum is 0x%02X\n", cs);
							outbuffer[len]=cs;
							FlushFileBuffers(h);
							WriteFile(h, &outbuffer, len+1, &dwBytesCount, NULL);
							break;
				case 'w':	// backwards compatible w, used by older tunerpro, but same as W
				case 'W':	len=packetbuffer[1]==0?256:packetbuffer[1];
							address=(packetbuffer[2]<<8)+packetbuffer[3];
							printf("Write request, %d bytes to 0x%04X\n", len, address);
							for (i=0;i<len;i++)	{
								//printf("command thread says: copying byte %d to 0x%4X\n", i, address+i);
								// only copy and flag if different
								if (emurom[address+i] != packetbuffer[4+i])	{
									printf("update 0x%02X to 0x%02X\n",emurom[address+i], packetbuffer[4+i] );
									emurom[address+i]=packetbuffer[4+i];
									emurom[address-0x8000+i]=1;
								}
							}
							printf("Wrote OK to tunerpro return code=%d ",  WriteFile(hSerialTP, okay, 1, &dwBytesCount, NULL));
							printf("bytes=%d\n", dwBytesCount);
							break;
				case 'K':	WaitForSingleObject( ghMutex, INFINITE);
							printf("ALDL passthrough\n");
							for (i=0; i<dwBytesCount; i++)	{
								printf("%02X ",packetbuffer[i]);
							}
							printf("\n");
							len = packetbuffer[2];
							FlushFileBuffers(hSerialECU);
							WriteFile(hSerialECU, &packetbuffer[3], packetbuffer[1], &dwBytesCount, NULL);
							// echo test WriteFile(hSerialTP, &packetbuffer[3], packetbuffer[1], &dwBytesCount, NULL);
							if (ReadFile(hSerialECU, packetbuffer, sizeof(packetbuffer), &dwBytesCount, NULL))	{
								if (dwBytesCount != len)	{
									printf("read too many bytes back from ecu (got %d, expecting %d), ignoring response\n", dwBytesCount, len);
								}
								printf("ECU Packet:\n");
								for (cs=0, i=0; i<dwBytesCount; i++)	{
									printf("%02X ", packetbuffer[i]);
								}
								printf("\n");
								memset(outbuffer, 0, sizeof(outbuffer));							
								memcpy(outbuffer, packetbuffer, dwBytesCount);
								// packet length + 6 bytes for 3x 2 byte A/D channels
								for (cs=0, i=0; i<len+6; i++)	{
									//printf("cs=%02X adding %02X and answer=0x%02X & FF\n", cs, outbuffer[i], cs+outbuffer[i] & 0xFF);
									cs=(cs+outbuffer[i]) & 0xFF;
								}
								printf ("adding checksum to position %d 0x%02X\n", len+6, cs);
								outbuffer[len+6]=cs;
								printf ("replying to tunerpro with\n");
								for (i=0; i<len+7; i++)	{
									printf("%02X ",outbuffer[i]);
								}
								WriteFile(h, &outbuffer, len+7, &dwBytesCount, NULL);	
							}
							printf("\n\n");
							ReleaseMutex(ghMutex);
							break;
				case 0xF4:	WaitForSingleObject( ghMutex, INFINITE);
							printf("direct ALDL passthrough\n");
							WriteFile(hSerialECU, packetbuffer, len, &dwBytesCount, NULL);
							printf("sent %d bytes to the ecu\n", dwBytesCount);
							WriteFile(h, packetbuffer, len, &dwBytesCount, NULL);
							printf("sent %d bytes of echo back to sender\n", dwBytesCount);
							// need to wait a little longer for the 60 byte packet
							Sleep(100);
							ReadFile(hSerialECU, packetbuffer, sizeof(packetbuffer), &dwBytesCount, NULL);
							printf("read %d bytes of reply\n", dwBytesCount);
							WriteFile(h, packetbuffer, dwBytesCount, &dwBytesCount, NULL);
							printf("sent %d bytes of replay back as answer\n", dwBytesCount);
							for (i=0; i<dwBytesCount; i++)	{
								printf("%02X ",packetbuffer[i]);
							}
							printf("\n\n");
							ReleaseMutex(ghMutex);
							break;
				default:	printf("Unknown command: ");
							for (i=0; i<dwBytesCount; i++)	{
								printf("%02X ",packetbuffer[i]);
							}
							printf("\n\n");
			}
		FlushFileBuffers(h);
		}
		else {
			//printf ("cmd thread loop\n");
			Sleep(20);
			FlushFileBuffers(h);
		}
	}

	return 1;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// thread stuff
	HANDLE hWriteThread;
	HANDLE hInputThreadTP;
	HANDLE hInputThreadSP;
	DWORD dwGenericThread;

	printf("Welcom to antusprom! Your realtime ALDL solution! :)\n");
	printf("V%01d.%01d\n\n", VMAJ, VMIN);
	printf("(c) 2009 Antus ant@symons.net.au\n\n");
	
	if (echo == 1)	{
		printf("This version built to expect ALDL cable echo\n");
	}
	
	if (argc < 3)	{
		printf("Usage: %S COM1 COM2 <COM3>\n\nwhere COM1 is the com port with ECU connected (can be invalid for no ecu)\nand COM2 is a com0com loopback serial.\nCOM3 is another com0com loopback\n\n", argv[0]);
		exit(1);
	}

	setup_ecu_comm(argv[1]);
	setup_other_comm(argv[2], hSerialTP);
	printf ("tunerproRT link com port init OK\n\n");
	if (argv[3])	{
		setup_other_comm(argv[3], hSerialSP);
		printf ("forwarding com port init OK\n\n");
	}

	ghMutex = CreateMutex( NULL, FALSE, NULL);   
    
	if (ghMutex == NULL) {
        printf("CreateMutex error: %d\n", GetLastError());
        exit(12);
    }

	if (ecu)	{
		printf ("starting writer thread\n");
		hWriteThread = CreateThread(NULL, 0, WriteThread,&emurom, 0, &dwGenericThread);
		if(hWriteThread == NULL)	{
				DWORD dwError = GetLastError();
				printf("Error creating writer thread %d\n\n", dwError);
				exit(10);
		}
	}

	printf ("starting TP input thread\n");
	hInputThreadTP = CreateThread(NULL, 0, input_handler_thread, hSerialTP, 0, &dwGenericThread);
	if(hInputThreadTP == NULL)	{
			DWORD dwError = GetLastError();
			printf("Error creating TP thread %d\n\n", dwError);
			exit(10);
	}
	

	if (argv[3])	{
		printf ("starting ALDL forwarding thread\n");
		hInputThreadSP = CreateThread(NULL,0, input_handler_thread, hSerialSP, 0, &dwGenericThread);
		if(hInputThreadSP == NULL)	{
				DWORD dwError = GetLastError();
				printf("Error creating SP thread %d\n\n", dwError);
				exit(10);
		}
	}

	while(1)	{
		Sleep(10000);
	}

	CloseHandle(ghMutex);
	if (hSerialECU) CloseHandle(hSerialECU);
	if (hSerialTP) CloseHandle(hSerialTP);
	if (hSerialSP) CloseHandle(hSerialTP);

	return 0;
}

